/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__PYTHON__)
  #include <Python.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(__MYSQL__)
  #include "KMySql.h"
#endif

#include "KCanElster.h"

#include "KElsterTable.h"


void KSniffFrame::SetValue(unsigned Value)
{
  if (3 <= Len && Len <= 5)
  {
    Data[Len] = (unsigned char) (Value >> 8);
    Data[Len+1] = (unsigned char) (Value & 0xff);  
  }
}

unsigned KSniffFrame::GetValue() const
{
  if (3 <= Len && Len <= 5)
  {
    return 256*Data[Len] + Data[Len + 1];
  }
  return 0;
}


////////////////////////////////////////////////////////////////////////////////

KCanElster::KCanElster()
{
  Trace = false;
  SendCanId = 0x700;
  
  for (int i = 0; i <= (int) High(Sniffed); i++)
    Sniffed[i] = NULL;
  
  for (int i = 0; i <= (int) High(UsedCanId); i++)
    UsedCanId[i] = false;
}

KCanElster::~KCanElster()
{
  for (int i = 0; i <= (int) High(Sniffed); i++)
  {
    while (Sniffed[i])
    {
      KSniffFrame * SF = Sniffed[i]->Next;
      
      delete Sniffed[i];
      Sniffed[i] = SF;
    }
  }
}

bool KCanElster::Init(const char * CanDev, unsigned short DefaultId)
{
  Trace = false;
  SendCanId = DefaultId;
  
  return KCanServer::Init(CanDev);
}

bool KCanElster::AddToSniffed(const KCanFrame & Frame)
{
  unsigned char Index = Frame.Data[0];
  
  if ((Frame.Len != 3 && Frame.Len != 5) ||
      (Index & 0xf) == 1)
    return false;
  
  if (SearchSniffFrame(Frame))
    return false;
  
  KSniffFrame * SF = new KSniffFrame;
  
  SF->Len = Frame.Len;
  for (int i = 0; i < Frame.Len; i++)
    SF->Data[i] = Frame.Data[i];
  
  SF->SetValue(0x8000); // Wert ist ungültig
  
  SF->Next = Sniffed[Frame.Data[0]];
  Sniffed[Frame.Data[0]] = SF;
  
  return true;  
}

KSniffFrame * KCanElster::SearchSniffFrame(const KCanFrame & Frame)
{
  KSniffFrame * SF = Sniffed[Frame.Data[0]];
  
  while (SF)
  {
    if (Frame.Len >= SF->Len)
    {
      for (int i = SF->Len-1; i >= 0; i--)
      {
        if (SF->Data[i] != Frame.Data[i])
          break;
        
        if (i <= 0)
          return SF;
      }
    }
    SF = SF->Next;
  }
  return NULL;
}

void KCanElster::PrintFrame(const KCanFrame & Frame)
{
  char Buffer[256];
  
  NCanUtils::FormatFrame(Frame, false, Buffer);
  printf("%s", Buffer);
}

void KCanElster::PrintSendFrame()
{
  PrintFrame(SendFrame);
}

void KCanElster::PrintRecvFrame()
{
  PrintFrame(RecvFrame);
}

bool KCanElster::Send_Frame()
{
  bool Ok = /* SendDriver-> */ SendData(SendFrame);
  
  if (Trace)
    PrintSendFrame();
  
  return Ok;
}

bool KCanElster::SniffFrame()
{
  bool Ok = GetFrame(RecvFrame);
  
  if (Ok)
  {
    KSniffFrame * SF = SearchSniffFrame(RecvFrame);
    if (SF &&
        RecvFrame.Len >= SF->Len + 2)
    {
      int len = SF->Len;
      SF->Data[len] = RecvFrame.Data[len];
      SF->Data[len+1] = RecvFrame.Data[len+1];
      SF->TimeStampDay = RecvFrame.TimeStampDay;
      SF->TimeStampMs = RecvFrame.TimeStampMs;
    }
  }
  return Ok;
}

void KCanElster::EmptyServer()
{
  bool Ok;
  do
  {
    Ok = SniffFrame();
    if (Ok && Trace &&
        SearchSniffFrame(RecvFrame))
      PrintRecvFrame();
  } while (Ok);
}

bool KCanElster::Get_Frame()
{
  bool Ok = SniffFrame();
  
  if (Ok && Trace)
    PrintRecvFrame();
  
  return Ok;
}

bool KCanElster::Send(unsigned count, bool WaitForAnswer)
{
  for (int i = 0; (unsigned) i < count; i++)  // 3 Versuche
  {
    EmptyServer();
    
    if (Send_Frame())
    {
      NCanUtils::SleepMs(20); // für den Raspi genügen 10 ms
      while (Get_Frame())
      {
        if (RecvFrame.EqualData(SendFrame))
        {
          if (!WaitForAnswer)
            return true;
          
          for (int k = 0; k < 50; k++)
          {
            NCanUtils::SleepMs(1);
            if (Get_Frame() &&
                RecvFrame.IsAnswer(SendFrame))
            {
              return true;
            }
          }
        }
      }
    }
    if ((unsigned)(i+1) < count)
      NCanUtils::SleepMs(1000);
  }
  
  return false;
}

void KCanElster::InitSendFrame(unsigned short receiver_id, unsigned short TelegrammId)
{
  SendFrame.InitElsterFrame(SendCanId, receiver_id, TelegrammId);
}

bool KCanElster::GetValue(unsigned short receiver_id, unsigned short TelegrammId, unsigned short & Value)
{
  InitSendFrame(receiver_id, TelegrammId);
  
  if (Send() && RecvFrame.Len == 7)
  {
    Value = RecvFrame.Value();
    return true;
  }
  
  return false;
}

bool KCanElster::SetValue(unsigned short receiver_id, unsigned short TelegrammId, unsigned short Value)
{
  InitSendFrame(receiver_id, TelegrammId);
  
  SendFrame.Len = 7;
  SendFrame.Data[0]++;
  SendFrame.Data[5] = (unsigned char)(Value >> 8);
  SendFrame.Data[6] = (unsigned char)(Value & 0xff);
  
  return SetValue();
}

bool KCanElster::SetValue()
{
  if ((SendFrame.Data[0] & 0x0f) != 2)
    return false;
  
  if (!Send(1, false))
    return false;
  
  NCanUtils::SleepMs(50);
  SendFrame.Data[0]--;
  if (!Send())
    return false;

  bool Ok = RecvFrame.Changed(SendFrame);
  
  SendFrame.Data[0]++;
  
  return Ok;
}

void KCanElster::GetValidDevices()
{
  printf("list of valid can id's:\n\n");
  for (int recv = 0; recv < 0x10; recv++)
    if (recv != SendCanId / 0x80)
    {
      int count = 0;
      for (int i = 0; i < 0x10; i++)
      {
        unsigned short Value;
        if (GetValue(0x80*recv + i, 0x000b, Value) &&
            (recv != 0 || Value != 0))
        {
          count++;
          printf("  %3.3x (%4.4x = %d)\n",
                 0x80*recv + i, Value, ((Value & 0xff) << 8) + (Value >> 8));
          UsedCanId[16*recv + i] = true;
        }
        if (count < i-2)
          break;
      }
    }
  printf("\n");
}

void KCanElster::Scanner(unsigned short receiver_id, bool get_all)
{
  unsigned short Value;
  char Buffer[256];
 
  for (unsigned Pattern = 0; Pattern < 0x10000; Pattern++)
  {
    const ElsterIndex * Elster = GetElsterIndex(Pattern);
    
    if (Pattern > 0x0200 && !Elster)
    {
      
    } else
    if (GetValue(receiver_id, Pattern, Value) &&
        Value != 0x8000)
    {
      if (get_all)
      {
        printf("  { 0x%3.3x, 0x%4.4x, 0x%4.4x},\n", receiver_id, Pattern, Value);
      } else
      if (Value || !(receiver_id & 7))
      {        
        if (!Elster ||
            Value != 0x8000)
        {
          sprintf(Buffer, "%4.4x:  %4.4x", Pattern, Value);
          if (Elster)
          {
            if (Value != 0x8000)
            {
              char Val[32];
          
              SetValueType(Val, Elster->Type, Value);
              sprintf(Buffer + strlen(Buffer), "  (%s  %s)", Elster->Name, Val);
            } else
              sprintf(Buffer + strlen(Buffer), "  (%s)", Elster->Name);
          }
          printf("%s\n", Buffer);
        }
      }
    }
  }  
}

bool KCanElster::GetInitCS_Id()
{
  unsigned short Value;
  
  bool Ok = GetValue(0x601, 0x000b, Value);
  if (!Ok)
    GetValue(0x301, 0x000b, Value);
  
  // http://www.ip-symcon.de/forum/threads/15888-Wolf-CSZ-mit-ISM1
  // eventuell mit direkter RS232-Schnittststelle an WP
  if (!Ok)
    Ok = GetValue(0x0000, 0x015b, Value); // GERAETEKONFIGURATION_2
  
  if (Ok)
  {
    KComfortFrame::InitId = (Value >> 8) + (Value << 8);
    return true;
  } 
  return false;
}

#if defined(__SCAN__)

void PrintComfortSoftFormat(const KCanFrame & Frame)
{
  char str[128];
  KCanFrame Frame1;
  KComfortFrame cf;
  KCan232Frame can232;
  
  cf.GetCanFrame(Frame);
  cf.SetToString(str);
  can232.GetCanFrame(Frame);
  printf("ComfortSoft: %s   can232: %s\n", str, can232.msg);
  cf.SetCanFrame(Frame1);
  
  bool Equal = Frame.Id == Frame1.Id;
  for (int i = 0; i < Frame.Len; i++)
    if (Frame.Data[i] != Frame1.Data[i])
      Equal = false;
  if (!Equal)
  {
    NCanUtils::FormatFrame(Frame1, false, str);
    printf("invalid back translated telegram: %s\n", str);
  }
}

int main(int argc, char* argv[])
{
  KCanElster Can;

#if defined(__WINDOWS__)
  printf("elster-kromschroeder can-bus address scanner and test utility\n"
         "copyright (c) 2014 Juerg Mueller, CH-5524\n\n");
#else
  printf("elster-kromschröder can-bus address scanner and test utility\n"
         "copyright (c) 2014 Jürg Müller, CH-5524\n\n");
#endif  
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  bool params_ok = argc == 3 || argc == 4;
  bool scan_only = false;
  bool write = false;
  unsigned short Index = 0;
  unsigned short Value = 0;
  unsigned short Value1 = 0;
  unsigned short Value2 = 0;
  const char * Line;
  const char * last_pos;  
  unsigned sender_can_id = 0;
  unsigned scan_can_id = 0;
  
  if (params_ok)
  {
    Line = argv[2];
    sender_can_id = NCanUtils::GetHex(Line);
    switch (sender_can_id & ~3)
    {
      case 0x680:
      case 0x700:
      case 0x780:
        break;
      default:
        params_ok = false;
        break;
    }
  }
  
  bool get_total = false;  
  if (params_ok && argc == 4)
  {
    get_total = !strcmp("total", argv[3]);
  }
  if (params_ok && argc == 4 && !get_total)
  {
    Line = argv[3];
    last_pos = Line;
    // 000.0000 oder 000.0000.0000
    scan_can_id = NCanUtils::GetHex(Line);
    if (Line - last_pos != 3 ||
        (scan_can_id < 0x100 && scan_can_id > 7) ||
        scan_can_id > 0x687 ||
        (scan_can_id & 0x078))
    {
      printf("\n%x is not a legal can id!\n\n", scan_can_id);
      params_ok = false;
    }
  
    last_pos = Line;
    scan_only = *Line != '.';
    if (!scan_only)
    {
      Index = NCanUtils::GetHex(++Line);
      if (Line - last_pos != 5)
        params_ok = false;
    }
    last_pos = Line;
    write = *Line == '.';
    if (write)
    {
      Value = NCanUtils::GetHex(++Line);
      if (Line - last_pos != 5 ||
          sender_can_id == 0)
        params_ok = false;
    }
  }
  if (!params_ok)
  {
    printf("use scanner\n"
           "can_scan <can dev> <sender can id> <scan can id> [ \"trace\" ]\n"
           "  or read/change value\n"
           "can_scan <can dev> <sender can id> <receiver can id> \".\" <elster index> [ \".\" <new value> ] [ \"trace\" ]\n\n"
           "example: can_scan can0 780 180           (scan all elster indices from 0000 to 1fff)\n"
           "or       can_scan can0 780 180.0126      (read minutes at elster index 0126)\n"
           "or       can_scan can0 780 180.0126.0f00 (set minutes to 15)\n\n"
           "allowed sender can id: 680 or 700 or 780\n\n"
           "BE CAREFUL WITH THIS TOOL !!!\n\n");
    return 0;
  }

  if (!Can.Init(argv[1], sender_can_id))
  {
    printf("device \"%s\" not connected\n", argv[1]);
    return -1;
  }
  
  Can.Trace = trace;
  Can.Resume();
  
  if (get_total)
  {
    Can.GetValidDevices();
    printf("\n");
    for (int i = 0; i <= (int) High(Can.UsedCanId); i++)
      if (Can.UsedCanId[i])
        Can.Scanner(8*(i & 0xf0) + (i & 0xf), true);
  } else
  if (argc == 3)
  {
    Can.GetValidDevices();
  } else
  if (scan_only)
  {
    Can.Scanner(scan_can_id, false);
  } else {
    bool Ok;
  
    if (write)
      Ok = Can.SetValue(scan_can_id, Index, Value);
    else 
      Ok = Can.GetValue(scan_can_id, Index, Value);
    
    if (Ok && trace)
    {
      printf("\nsend ");
      PrintComfortSoftFormat(Can.SendFrame);
      printf("recv ");
      PrintComfortSoftFormat(Can.RecvFrame);
      printf("\n");
      
    }
    
    if (!Ok)
    {
      if (write)
        printf("value not set\n");
      else
        printf("no value read\n");
    } else {
      if (write)
        printf("value set\n");
      
      printf("value: %4.4x", Value);

      const ElsterIndex * elst_ind = GetElsterIndex(Index);
      if (elst_ind)
      {
        bool double_val = false;
        char Val[32];

        if (!write && (Value & 0xff00) != 0x8000)
        {
          if (elst_ind->Type == et_double_val ||
              elst_ind->Type == et_triple_val)
          {
            double_val = Can.GetValue(scan_can_id, ++Index, Value1);
            if (double_val && elst_ind->Type == et_triple_val)
              double_val = Can.GetValue(scan_can_id, ++Index, Value2);
          }
        }
        if (double_val)
        {
          const ElsterIndex * elst_ind1 = GetElsterIndex(Index);
          if (elst_ind1)
            elst_ind = elst_ind1;
          
          double d = (double)(Value) / 1000.0 + (double)(Value1);
          if (elst_ind->Type == et_triple_val)
          {
            d /= 1000.0;
            d += (double)(Value2);
            sprintf(Val, "%.6f", d);
          } else
            sprintf(Val, "%.3f", d);
        } else
          SetValueType(Val, elst_ind->Type, Value);
        printf("  (%s  %s)", elst_ind->Name, Val);
      }
      printf("\n");    
    }
  }
  Can.Halt();
  
  return 0;  
}

#endif

#if defined(__VZ__)

#include "XmlParser.h"
#include "KTcpCanClient.h"

int main(int argc, char* argv[])
{ 
  printf("Volkszaehler can-bus logger\n"
         "copyright (c) 2014 Jürg Müller, CH-5524\n\n");

  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  if (argc != 2)
  {
    printf("usage:\n"
           "can_vz <xml configuration> [ \"trace\" ]\n"
           "example: can_vz VZ_config.xml\n\n");
    return 0;
  }
  
  KParser Parser;
  KXmlNode Root;
  const KXmlNode * Node;
  char buffer[2048];
  char query_template[2048];
  
  if (!Parser.ParseFile(argv[1], &Root))
    return -1;
  
  const char * device = Root.FindNodeAttr("device", "Name");
  const char * sender = Root.FindNodeAttr("sender", "Name");
  const char * template_ = Root.FindNodeAttr("template", "Name");
  const char * ip_address = Root.FindNodeAttr("ip_address", "Name");
  const char * secs = Root.FindNodeAttr("sleep", "Name");
  unsigned sender_id = 0x700;
#if defined(__MYSQL__)
  bool use_mysql = Root.Find("mysql") != NULL;
  KMySql MySql;
#else
  bool use_mysql = false;
#endif
  int seconds = 104;
  
  if (!use_mysql &&
      (!template_ || strlen(template_) == 0))
  {
    printf("template is missing\n");
    return -1;
  }
  if (!use_mysql && (!ip_address || strlen(ip_address) == 0))
  {
    printf("IP address is missing\n");
    return -1;
  }
  
  if (sender)
  {
    unsigned id = NCanUtils::GetHex(sender);
    switch (id)
    {
      case 0x580:
      case 0x680:
      case 0x700:
      case 0x780:
        sender_id = id;
        break;
      default:
        break;
    }
  }
  
#if defined(__MYSQL__)
  if (use_mysql &&
      !MySql.GetXmlParams(&Root))
    return -1;
#endif
  
  if (secs)
  {
    int sec;
    if (NCanUtils::GetInt(secs, sec) &&
        !*secs &&
        60 <= sec && sec <= 3600)
      seconds = sec;
  }
  if (trace)
  {
    printf("device: %s\n", device);
    printf("sender: %x\n", sender_id);
    printf("sleep: %d secs\n", seconds);
    if (!use_mysql)
      printf("IP address: %s\n", ip_address);
    printf("\n");
  }
  
  // cannel check
  int cnt = 1;
  Node = NULL;
  while (Root.GetNextNode(Node))
  {
    if (strlen(Node->mName) == 0)
      return -1;
    
    if (!strcmp(Node->mName, "channel"))
    {
      const char * can_msg = Node->FindAttr("can_msg");
      const char * uuid = Node->FindAttr("uuid");
      const char * query = Node->FindAttr("query");
      const char * name = Node->FindAttr("Name");
      
      if (template_ && query &&
          strlen(template_) < High(query_template) + High(query))
      {
        sprintf(query_template, template_, query);
        query = query_template;
      }
      
      if (!name || strlen(name) == 0)
      {
        printf("\"Name\" is missing in channel %d\n", cnt);
        return -1;
      }
      if (use_mysql || query)
      {
        if (!query || strlen(query) == 0)
        {
          printf("query is missing in channel %d (%s)\n", cnt, name);
          return -1;
        }
        if (!strstr(query, "${value}"))
        {
          printf("${value} in query is missing %s (%s)\n", query, name);
          return -1;
        }
      } else {
        if (!uuid || strlen(uuid) == 0)
        {
          printf("uuid is missing in channel %d (%s)\n", cnt, name);
          return -1;
        }
      }
      if (!can_msg || strlen(can_msg) == 0)
      {
        printf("can_msg is missing in channel %d (%s)\n", cnt, name);
        return -1;
      } else {
        KCanFrame Frame;
        
        if (!Frame.GetDataFromStream(can_msg))
        {
          printf("invalid can_msg(%s): \"%s\"\n", name, can_msg);
          return -1;
        }
      }
      cnt++;
    }
  }  
  
  KCanElster Can;
  bool added = false;
  
  Node = NULL;
  while (Root.GetNextNode(Node))
  {
    if (!strcmp(Node->mName, "channel"))
    {
      KCanFrame Frame;
      const char * can_msg = Node->FindAttr("can_msg");
      
      Frame.GetDataFromStream(can_msg);
      if (Can.SearchSniffFrame(Frame))
      {
        printf("can_msg \"%s\" already exists\n", can_msg);
      }
      if (Can.AddToSniffed(Frame) && trace)
      {
        printf("frame can_msg added: %s\n", can_msg);
        added = true;
      }
    }
  }
  if (added)
    printf("\n");
  
  KTcpCanClient TcpClient;
  time_t timer;
  time_t new_timer;

  TcpClient.Trace = trace;
  TcpClient.SetBlockMode(bmBlocking);
  TcpClient.SetRemoteHost(ip_address);
  TcpClient.SetRemotePort(80);
  
  if (!Can.Init(device, sender_id))
  {
    printf("device \"%s\" not connected\n", Can.GetDev());
   
    Can.Halt();
    
    return -1;
  }

  Can.Trace = trace;
  Can.Resume();
  
  NCanUtils::SleepMs(10000);
  while (!Can.Terminated())
  {
    Node = NULL;
    time(&timer);
    while (Root.GetNextNode(Node))
    {
      if (!strcmp(Node->mName, "channel"))
      {
        const char * can_msg = Node->FindAttr("can_msg");
        
        const_cast<KXmlNode *>(Node)->can_value = 0x8000;

        Can.SendFrame.Init();
        Can.SendFrame.Id = Can.SendCanId;
        Can.SendFrame.GetDataFromStream(can_msg);
        if ((Can.SendFrame.Data[0] & 0xf) != 1)
        {
          Can.EmptyServer();
          
          KSniffFrame * SF = Can.SearchSniffFrame(Can.SendFrame);
          if (SF)
          {
            const_cast<KXmlNode *>(Node)->can_value = SF->GetValue();
            const_cast<KXmlNode *>(Node)->time_days = SF->TimeStampDay;
            const_cast<KXmlNode *>(Node)->time_ms = SF->TimeStampMs;
  
            SF->SetValue(0x8000);
          }
        } else
        if (Can.Send())
        {
          const_cast<KXmlNode *>(Node)->can_value = Can.RecvFrame.Value();
          const_cast<KXmlNode *>(Node)->time_days = Can.RecvFrame.TimeStampDay;
          const_cast<KXmlNode *>(Node)->time_ms = Can.RecvFrame.TimeStampMs;
        }
      }
    }
    
  #if defined(__MYSQL__)
    if (use_mysql)
    {
      MySql.Open();
    }
  #endif
    Node = NULL;
    while (Root.GetNextNode(Node))
    {
      if (!strcmp(Node->mName, "channel"))
      {
        if (Node->can_value != 0x8000)
        {
          const char * format = Node->FindAttr("format");
          const char * query = Node->FindAttr("query");
          ElsterType type = GetElsterType(format);
          char Value[64];
          SetValueType(Value, type, Node->can_value);
          
          if (query)
          {
            if (template_)
            {
              sprintf(query_template, template_, query);
              query = query_template;
            }            
            buffer[0] = 0;
            while (query && *query)
            {
              const char * append_value;
              const char * param = strstr(query, "${");
              const char * param_end = param ? strstr(param, "}") : NULL;
              int param_len = param && param_end ? (int)(param_end - param + 1) : 0;
              
              // Platzhalter ${...} durch den entsprecheden Wert ersetzen
              //
              // intern verwendete Platzhalter: ${value}, ${ts} und ${timestamp}
              //
              // alle anderen Platzhalter werden in der Parameterliste zum "channel" gesucht
              // z.B. für ${uuid}
              //
              // <channel uuid="xxx..."    ... />
              if (param && param_len)
              {
                char param_value[256];
                int n = (int)(param - query);
                int len = (int) strlen(buffer);
                strncpy(buffer + len, query, n);
                buffer[len + n] = 0;
                query += n;
                append_value = param_value;
                
                param_value[0] = 0;
                
                if (strstr(param, "${value}") == param)
                {
                  strcpy(param_value, Value);
                } else
                if (strstr(param, "${ts}") == param  ||
                    strstr(param, "${timestamp}") == param)
                {
                  if (param[3] == 'i')
                    NCanUtils::PrintMs(Node->time_days, Node->time_ms, param_value);
                  else
                    NCanUtils::PrintDateTime(Node->time_days, Node->time_ms, param_value);
                } else {
                  char param_name[64]; // z.B. ${uuid}
                  
                  int len = param_len - 3;
                  if (len >= (int) High(param_name))
                  {
                    len = High(param_name)-1;
                    if (trace)
                      printf("error in length of parameter\n");
                  }
                  strncpy(param_name, param + 2, len);
                  param_name[len] = 0;
                  
                  const char * found_param_value = Node->FindAttr(param_name);
                  if (!found_param_value ||
                      strlen(found_param_value) >= High(param_value))
                  {
                    if (trace)
                      printf("parameter value \"%s\" too long\n", found_param_value);
                  } else
                    strcpy(param_value, found_param_value);
                }                
                query += param_len;
              } else {
                append_value = query;
                query = NULL;
              }
              if (strlen(buffer) + strlen(append_value) < High(buffer))
                strcat(buffer, append_value);
              else
              if (trace)
                printf("buffer overflow\n");              
            }
          } else  {
            const char * uuid = Node->FindAttr("uuid");
            sprintf(buffer, template_, uuid, Value);
          }
          
        #if defined(__MYSQL__)
          if (use_mysql)
          {
            if (!MySql.Query(buffer))
              printf("error in query %s\n", buffer);
            else
            if (trace)
              printf("query: %s\n", buffer);
          } else
        #endif
          {
            if (trace)
            {
              strcat(buffer,
                     "\r\nAccept: text/plain, text/html; charset=UTF-8"
                     "\r\nAccept-Charset: ISO-8859-1"
                     "\r\nContent-Language: de");
            }
            int str_len;
            if (!TcpClient.ReadHttpStream(*(char **) NULL, str_len, buffer))
            {
              printf("TCP/IP server not connected\n");
              if (!trace)
                printf("==> %s\n", buffer);
            }
          }
          const_cast<KXmlNode *>(Node)->can_value = 0x8000;            
        } else
        if (trace)
        {
          const char * name = Node->FindAttr("Name");
          printf("\"%s\" has non or invalid value\n", name);
        }
      }
    }
    
  #if defined(__MYSQL__)
    if (use_mysql)
    {
      MySql.Close();
    }
  #endif
    
    time(&new_timer);
    int del = seconds - (int)(new_timer - timer);
    if (del < 45)
      del = 45;
    while (del > 0)
    {
      int d = del < 10 ? del : 10;
      
      NCanUtils::SleepMs(d*1000);
      del = del < 10 ? 0 : del - 10;
      
      Can.EmptyServer();
    }
  }
  
  return 0;
}

#endif

#if defined(__SIMULATION__)

#include "KScanTable.h"

int main(int argc, char* argv[])
{
  printf("CAN simulator\n"
         "copyright (c) 2014 Juerg Mueller, CH-5524\n\n");
  
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;

  KCanElster CanElster;
  KCanFrame Frame;
  KComfortFrame cf;
  KComfortFrame cfRes;
  char str[256];
  bool start_ok = argc == 2 || argc == 3;
  
  if (argc == 3)
  {
    if (!LoadScanTable(argv[2]))
    {
      printf("\ncould scan table \"%s\" not load\n", argv[2]);
      start_ok = false;
    }
  } else
    Init_scan_table();
  
  if (!start_ok)
  {
    printf("\nusage:\n"
#if defined(__WINDOWS__)
           "can_simul <com dev> [ <scan table name> ] [ \"trace\" ]\n"
           "example: can_simul com3 ..\\scan_data.inc \n\n");
#else
           "can_simul <tty dev> [ <scan table name> ] [ \"trace\" ]\n"
           "example: can_simul tty.usbserial-FTK1S17H ../scan_data.inc \n\n");
#endif
    return -1;
  }
  
  CanElster.driver_type = KCanServer::dt_cs;
  if (!CanElster.Init(argv[1], 0x680))
    return  -1;
  
  CanElster.Trace = trace;
  CanElster.Resume();
  NCanUtils::SleepMs(100);
  
  do
  {
    if (CanElster.GetFrame(Frame))
    {
      cf.GetCanFrame(Frame);
      
      if (trace)
      {
        cf.SetToString(str);
        printf("recv %s\n", str);
      }
      if (cf.IsInitFrame())
      {
        cfRes.SetOkValue(KComfortFrame::InitId);
      } else
      if (!Get_scan_value(cf, cfRes))
        cfRes.SetOk(false);

      if (trace &&
          cfRes.SetCanFrame(Frame) &&
          FormElsterTable(Frame, str))
      {
          printf("Elster %3.3x.%4.4x.%4.4x: %s\n", Frame.Id, Frame.Index(), Frame.Value(), str);
      }
      cfRes.SetToString(str);
      if (trace)
        printf("send %s\n", str);
      
      CanElster.SendData(cfRes);
    }
    NCanUtils::SleepMs(1);
  } while (!CanElster.Terminated());
  
  CanElster.Halt();
  
  return 0;
}

#endif

#if 0
//defined(__PYTHON__)

// Beispiel "spam" siehe
// 
// https://docs.python.org/2/extending/extending.html
//

#ifdef __cplusplus
extern "C" {
#endif
  
static KCanElster Can;

static PyObject * elsterError;

static bool InitCan(const char * dev)
{
  if (!Can.Init(dev, 0x700))
  {
    char err[128];
    sprintf(err, "device \"%s\" not connected\n", dev);
    
    PyErr_SetString(elsterError, err);
    return false;
  }
  Can.Trace = true;
  Can.Resume();
  
  return true;
}
  
static bool get_command(PyObject *args, const char * & command)
{
  if (!PyArg_ParseTuple(args, "s", &command))
    return false;
  
  if (!command)
    return false;
  
  while (*command == ' ')
    command++;
  
  return true;  
}

static bool get_unsigned_short(const char * & command, unsigned short & res)
{
  if (!NCanUtils::IsHexDigit(*command))
    return false;
  
  unsigned result = NCanUtils::GetHex(command);
  while (*command == ' ')
    command++;
  
  res = (unsigned short) result;
  
  return result <= 0xffff;  
}
  
static bool elsterStart(PyObject *args, unsigned short & send_id,
                        unsigned short & recv_id, unsigned short & elster_index)
{
  const char * command;

  if (!get_command(args, command))
    return false;
  
  if (!get_unsigned_short(command, send_id))
    return false;
  
  if (!get_unsigned_short(command, recv_id))
    return false;
  
  if (!get_unsigned_short(command, elster_index))
    return false;

  if (*command)
    return false;
  
  if (send_id != 0x680 && send_id != 0x700 && send_id != 0x780)
    return false;
  
  if (Can.SendCanId)
    InitCan("can0");
  
  Can.SendCanId = send_id;
  
  return true;    
}
  
  
////////////////////////////////////////////////////////////////////////////////
  
static PyObject * elster_getname(PyObject *self, PyObject *args)
{
  const char * command;
  unsigned short elster_index;
  
  if (!get_command(args, command))
    return NULL;
  
  if (!get_unsigned_short(command, elster_index))
      return NULL;
      
  if (*command)
    return NULL;
  
  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
    return PyBytes_FromString(elst_ind->Name);
  
  return PyLong_FromLong(-1);
}

static PyObject * elster_gettype(PyObject *self, PyObject *args)
{
  const char * command;
  unsigned short elster_index;

  if (!get_command(args, command))
    return NULL;
  
  if (!get_unsigned_short(command, elster_index))
    return NULL;

  if (*command)
    return NULL;
  
  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
    return PyLong_FromLong(int(elst_ind->Type));
  
  return PyLong_FromLong(-1);
}
  
static PyObject * elster_getvalue(PyObject *self, PyObject *args)
{
  unsigned short send_id;
  unsigned short recv_id;
  unsigned short elster_index;
  unsigned short value;
  
  if (!elsterStart(args, send_id, recv_id, elster_index))
    return NULL;
  
  if (!Can.GetValue((unsigned short) recv_id, elster_index, value))
    return NULL;
 
  return PyLong_FromLong(value);
}

static PyObject * elster_getstring(PyObject *self, PyObject *args)
{
  unsigned short send_id;
  unsigned short recv_id;
  unsigned short elster_index;
  unsigned short value;
  char Val[32];
  
  if (!elsterStart(args, send_id, recv_id, elster_index))
    return NULL;
  
  if (!Can.GetValue(recv_id, elster_index, value))
    return NULL;

  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
  {
    bool double_val = false;
    unsigned short value1;
    unsigned short value2;
    
    if ((value & 0xff00) != 0x8000)
    {
      if (elst_ind->Type == et_double_val ||
          elst_ind->Type == et_triple_val)
      {
        double_val = Can.GetValue(recv_id, ++elster_index, value1);
        if (double_val && elst_ind->Type == et_triple_val)
          double_val = Can.GetValue(recv_id, ++elster_index, value2);
      }
    }
    if (double_val)
    {
      const ElsterIndex * elst_ind1 = GetElsterIndex(elster_index);
      if (elst_ind1)
        elst_ind = elst_ind1;
      
      double d = (double)(value) / 1000.0 + (double)(value1);
      if (elst_ind->Type == et_triple_val)
      {
        d /= 1000.0;
        d += (double)(value2);
        sprintf(Val, "%.6f", d);
      } else
        sprintf(Val, "%.3f", d);
    } else
      SetValueType(Val, elst_ind->Type, value);
  } else
    sprintf(Val, "%d", value);
  
  return PyBytes_FromString(Val);
}
  
static PyObject * elster_system(PyObject *self, PyObject *args)
{
  const char *command;
  int sts;
  
  if (!PyArg_ParseTuple(args, "s", &command))
    return NULL;
  
  sts = system(command);
  if (sts < 0)
  {
    PyErr_SetString(elsterError, "System command failed");
    return NULL;
  }

  return PyLong_FromLong(sts);
}

static PyMethodDef elsterMethods[] =
{
  {"system",    elster_system,    METH_VARARGS, "Execute a shell command."},
  {"getvalue",  elster_getvalue,  METH_VARARGS, "Get Elster value"},
  {"getstring", elster_getstring, METH_VARARGS, "Get Elster string"},
  {"getname",   elster_getname,   METH_VARARGS, "Get Elster name"},
  {"gettype",   elster_gettype,   METH_VARARGS, "Get Elster type"},
  {NULL, NULL, 0, NULL}
};

////////////////////////////////////////////////////////////////////////////////
  
#if PY_MAJOR_VERSION >= 3
  struct module_state
  {
    PyObject * error;
  };

  #define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
  
  static PyObject * error_out(PyObject *m)
  {
    struct module_state *st = GETSTATE(m);
    PyErr_SetString(st->error, "something bad happened");
    return NULL;
  }
  
  static int elster_traverse(PyObject *m, visitproc visit, void *arg)
  {
    Py_VISIT(GETSTATE(m)->error);
    return 0;
  }
  
  static int elster_clear(PyObject *m)
  {
    Py_CLEAR(GETSTATE(m)->error);
    if (Can.SendCanId)
      Can.Halt();
    
    return 0;
  }
  
  static struct PyModuleDef moduledef =
  {
    PyModuleDef_HEAD_INIT,
    "elster",
    NULL,
    sizeof(struct module_state),
    elsterMethods,
    NULL,
    elster_traverse,
    elster_clear,
    NULL
  };

  
  PyObject * PyInit_elster(void)
  {
    return PyModule_Create(&moduledef);
  }
#else
void initelster(void)
{
  (void) Py_InitModule("elster", elsterMethods);
}
#endif

int main(int argc, char *argv[])
{
  Can.SendCanId = 0;
  /* Pass argv[0] to the Python interpreter */
  #if PY_MAJOR_VERSION >= 3
    wchar_t name[1024];
    for (unsigned i = 0; i <= strlen(argv[0]); i++)
      name[i] = argv[0][i];
  
    Py_SetProgramName(name);
  #else
    Py_SetProgramName(argv[0]);
  #endif
  
  /* Initialize the Python interpreter.  Required. */
  Py_Initialize();
  
#if PY_MAJOR_VERSION < 3
  initelster();
#endif
  
  return 0;
}

#ifdef __cplusplus
}
#endif

#endif
